-- scope stuff
gearquipper = gearquipper or {};
local c = gearquipper;

----- [ general ] -----

-- character fields
local FIELD_CURRENTSET = "FIELD_CURRENTSET";
local FIELD_LASTSET = "FIELD_LASTSET";
local FIELD_SETS = "FIELD_SETS";
local FIELD_EVENTS = "FIELD_EVENTS";
local FIELD_INVENTORY = "FIELD_INVENTORY";

-- set fields
local FIELD_SHOWCLOAK = "FIELD_SHOWCLOAK";
local FIELD_SHOWHELMET = "FIELD_SHOWHELMET";
local FIELD_SLOTS = "FIELD_SLOTS";
local FIELD_SLOTSTATES = "FIELD_SLOTSTATES";
local FIELD_ACTIONSLOTS = "FIELD_ACTIONSLOTS";

local FIELD_OPT_ACTIONSLOTS = "FIELD_OPT_ACTIONSLOTS";
local FIELD_OPT_PARTIAL = "FIELD_OPT_PARTIAL";

-- misc fields
c.FIELD_ID = "FIELD_ID";
c.FIELD_TYPE = "FIELD_TYPE";
c.FIELD_SUBTYPE = "FIELD_SUBTYPE";
c.FIELD_SUBSUBTYPE = "FIELD_SUBSUBTYPE"; -- transient
c.FIELD_NAME = "FIELD_NAME";
c.FIELD_PVE = "FIELD_PVE";
c.FIELD_PVP = "FIELD_PVP";

function c:InitDatabase()
	GQ_DATA[c:GetRealmName()] = GQ_DATA[c:GetRealmName()] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()] = GQ_DATA[c:GetRealmName()][c:GetCharName()] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS] or {};
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY] or {};
end

function c:DataBase_Update14()
	for realmName in pairs(GQ_DATA) do
		for charName in pairs(GQ_DATA[realmName]) do
			for setName in pairs(GQ_DATA[realmName][charName][FIELD_SETS]) do
				GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES] = GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES] or {};
				for slotId in pairs(c:GetSlotInfo()) do
					if GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] == nil then
						GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] = true;
					end
				end
			end
		end
	end
end

function c:DataBase_Update17()
	for realmName in pairs(GQ_DATA) do
		for charName in pairs(GQ_DATA[realmName]) do
			GQ_DATA[realmName][charName][FIELD_EVENTS] = {};
		end
	end
end

----- [ sets ] -----

function c:LoadSetNames()
	local setNames = {};
	for setName in pairs(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS]) do
		table.insert(setNames, setName);
	end

	table.sort(setNames);
	return setNames;
end

function c:LoadCurrentSetName()
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET];
end

function c:LoadPreviousSetName()
	return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_LASTSET];
end

function c:SaveCurrentSetName(setName)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_LASTSET] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET];
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_CURRENTSET] = setName;
end

function c:SaveSet(setName, notify)
	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS] or {};
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] or {};

		for slotId, slotName in pairs(c:GetSlotInfo()) do
			local slotState = c:LoadSlotState(slotId, setName);
			if not c:LoadPartialOption(setName) or slotState == nil or slotState then
				c:SaveSlot(slotId, setName, notify);
			end
		end

		c:SaveCloakAndHelmet(setName, notify);
		c:SaveActionSlotsOption(setName, true);
		c:SaveActionConfiguration(setName, notify);
		c:SaveCurrentSetName(setName);

		c:RefreshSetList(setName);
		c:GetBindingEntries(true);
		c:SetPaperDollLabelText();
	end
end

function c:SaveDefaultSet()
	c:SaveSet(c:GetText("default"), false);
end

function c:DeleteSet(setName)
	GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = nil;

	if c:LoadCurrentSetName() == setName then
		c:SaveCurrentSetName(nil);
	end
	for eventName, bindingName in pairs(c:LoadEventBindings()) do
		if setName == bindingName then
			c:SaveEventBinding(eventName, nil);
		end
	end

	CloseDropDownMenus();
	c:RefreshSetList();
	c:GetBindingEntries(true);
	c:RefreshEventEntries();
	c:SetPaperDollLabelText();
	c:HideSlotStateBoxes();
end

----- [ slots ] -----

function c:SaveSlot(slotId, setName, notify, itemString)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	if setName then
		itemString = itemString or c:GetItemString(slotId);
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS][slotId] ~= itemString then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS][slotId] = itemString;
			if notify  then
				c:NotifyInChat(c:GetText("<%s> Saved slot %s (%s): %s", setName, slotId, c:GetDisplaySlotName(slotId), c:GetItemLink(itemString)));
			end
		end
	end
end

function c:LoadSlot(slotId, setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTS][slotId];
	end
end

function c:LoadSlotState(slotId, setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId];
	end
end

function c:SaveSlotState(slotId, active, notify, setName)
	setName = setName or c:LoadCurrentSetName();
	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES] or {};
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] ~= active then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SLOTSTATES][slotId] = active;
			if notify then
				c:NotifyInChat(c:GetText("<%s> Saved slot state for %s (%s): %s", setName, slotId, c:GetDisplaySlotName(slotId), c:BoolToText(active, c:GetText("active"), c:GetText("not active"))));
			end
		end
	end
end

function c:LevelUp(newLevel)
	-- itemStrings contain the level at which they were created, which makes an update neccessary if there has been a lvlup
	if newLevel then
		-- sets
		for setName in pairs(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS]) do
			for slotId, slotName in pairs(c:GetSlotInfo()) do
				local itemString = c:LoadSlot(slotId, setName);
				if not c:IsEmpty(itemString) then
					local stringValues = { strsplit(":", itemString) };
					stringValues[10] = newLevel;
					c:SaveSlot(slotId, setName, false, table.concat(stringValues, ":"));
				end
			end
		end

		-- last inventory locations
		local copy = c:Deepcopy(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY]);
		for itemString, values in pairs(copy) do
			-- remove old value
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][itemString] = nil;

			-- build new value
			local stringValues = { strsplit(":", itemString) };
			stringValues[10] = newLevel;
			local newItemString = table.concat(stringValues, ":");
			if newItemString then
				-- set new value
				GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][table.concat(stringValues, ":")] = values;
			end
		end
	end
end

local macros;
function c:GetMacros(refresh)
	if not macros or refresh then
		macros = {
			["all"] = {},
			["global"] = {},
			["char"] = {}
		}

		local numGlobal, numLocal = GetNumMacros(); -- 1-120: global; 121-138: local
		for id = 1, numGlobal do
			local name, iconTexture, body, isLocal = GetMacroInfo(id);
			macros["all"][id] = name;
			macros["global"][id] = name;
		end

		for id = 1, numLocal do
			id = id + 120;
			local name, iconTexture, body, isLocal = GetMacroInfo(id);
			macros["all"][id] = name;
			macros["char"][id] = name;
		end
	end
	return macros;
end

function c:ProcessMacroUpdate()
	-- update macro ids after macro has been deleted
	if c.macroAction then
		local oldMacrosAll, newMacrosAll, oldMacros, newMacros, realmName, charName = c:GetMacros(), c:GetMacros(true);
		if c.macroAction[c.FIELD_TYPE] == 1 then
			oldMacros = oldMacrosAll["global"];
			newMacros = newMacrosAll["global"];
		elseif c.macroAction[c.FIELD_TYPE] == 2 then
			oldMacros = oldMacrosAll["char"];
			newMacros = newMacrosAll["char"];
			realmName = c:GetRealmName();
			charName = c:GetCharName();
		end

		if oldMacros and newMacros then
			-- preparation
			local newIds = {};
			if c.macroAction[c.FIELD_NAME] == c.MACROACTION_EDIT then
				local renamedNewId = c:FindRenamedMacro(oldMacros, newMacros, realmName, charName);
				if renamedNewId then
					table.insert(newIds, renamedNewId);
				end
			elseif c.macroAction[c.FIELD_NAME] == c.MACROACTION_DELETE then
				c:UpdateMacro(c.macroAction[c.FIELD_ID], nil, realmName, charName);
			end

			for id, name in pairs(oldMacros) do
				local newId = GetMacroIndexByName(name);
				if newId and newId > 0 and id ~= newId then
					c:UpdateMacro(id, "R" ..  newId, realmName, charName);
					table.insert(newIds, newId);
				end
			end

			-- finish
			for _, id in ipairs(newIds) do
				c:UpdateMacro("R" .. id, id, realmName, charName);
			end
		end

		c.macroAction = nil;
	end
end

function c:FindRenamedMacro(oldMacros, newMacros, realmName, charName)
	for id, name in ipairs(oldMacros) do
		if not c:TableContains(newMacros, name) then
			-- renamed macro (old name)
			for newId, newName in ipairs(newMacros) do
				if not c:TableContains(oldMacros, newName) then
					-- renamed macro (new name)
					c:UpdateMacro(id, "R" .. newId, realmName, charName);
					return newId;
				end
			end
		end
	end
end

function c:UpdateMacro(oldMacroId, newMacroId, realmName, charName)
	if not realmName or not charName then
		-- global macros
		for realmName in pairs(GQ_DATA) do
			for charName in pairs(GQ_DATA[realmName]) do
				c:UpdateMacro(oldMacroId, newMacroId, realmName, charName);
			end
		end
	else
		-- local macro (character macro)
		for setName in pairs(GQ_DATA[realmName][charName][FIELD_SETS]) do
			for slotId, entry in ipairs(GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_ACTIONSLOTS]) do
				if entry and entry[c.FIELD_TYPE] == "macro" and entry[c.FIELD_ID] == oldMacroId then
					local old = GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][c.FIELD_ID];
					if newMacroId then
						GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][c.FIELD_ID] = newMacroId;
					else
						GQ_DATA[realmName][charName][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] = c.VALUE_NONE;
					end
				end
			end
		end
	end
end

----- [ actions ] -----

function c:SaveActionConfiguration(setName, notify)
	setName = setName or c:LoadCurrentSetName();
	if setName then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] or {};

		for slotId = 1, 120 do
			c:SaveActionSlot(slotId, setName, notify);
		end
	end
end

function c:SaveActionSlot(slotId, setName, notify, realmName, charName)
	slotId = tonumber(slotId);
	if slotId then
		realmName = realmName or c:GetRealmName();
		charName = charName or c:GetCharName();
		setName = setName or c:LoadCurrentSetName();
		if setName then
			if notify == nil then
				notify = true;
			end

			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] or {};
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS] or {};
			local entry = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId];

			if not c:IsSameAction(slotId, entry) then
				if HasAction(slotId) then
					local actionType, id, subType = GetActionInfo(slotId);

					local newEntry = {};
					newEntry[c.FIELD_TYPE] = actionType;
					newEntry[c.FIELD_ID] = id;
					newEntry[c.FIELD_SUBTYPE] = subType;

					GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] = newEntry;
				else
					GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] = c.VALUE_NONE;
				end

				if notify then
					local actionItemName = c:GetText("Empty");

					if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId] ~= c.VALUE_NONE then
						local aType, id = GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][c.FIELD_TYPE], GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS][slotId][c.FIELD_ID];

						if aType == "spell" then
							local name, rank = c:GetSpellName(id), c:GetSpellSubText(id);
							if rank and rank ~= "" then
								actionItemName = string.format("%s (%s)", name, rank);
							else
								actionItemName = name;
							end
						elseif aType == "item" then
							actionItemName = c:GetRarityColor(c:GetItemRarity(id))["hex"] .. c:GetItemName(id);
						elseif aType == "macro" then
							actionItemName = c:GetMacroName(id);
						end
					end
					c:NotifyInChat(c:GetText("<%s> Saved action slot %s: %s", setName, slotId, actionItemName));
				end
			end
		end
	end
end

function c:LoadActionSlots(setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_ACTIONSLOTS];
	end
end

----- [ event ] -----

function c:SaveEventBinding(binding, index)
	if index then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS][index] = binding;
	else
		if binding[c.FIELD_SUBSUBTYPE] then
			binding[c.FIELD_SUBTYPE] = binding[c.FIELD_SUBSUBTYPE];
			binding[c.FIELD_SUBSUBTYPE] = nil;
		end
		table.insert(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS], binding);
	end

	-- notification
	local event = c:GetEvents()[binding[c.FIELD_TYPE]];
	if event and binding[c.FIELD_SUBTYPE] then
		event = event .. " (" .. c:GetZoneDisplayName(binding[c.FIELD_SUBTYPE]["mapID"]) .. ")";
	end
	local setName = binding[c.FIELD_NAME];
	if setName == c.KEYWORD_PREVIOUS then
		setName = c:GetText("[Previous set]");
	end
	c:NotifyInChat(string.format("%s %s -> %s", c:GetText("Saved event:"), event, setName));
end

function c:LoadEventBindings(filter)
	if filter then
		local result = {};
		for index, binding in ipairs(c:LoadEventBindings()) do
			local applies = true;
			for filterName, filterValue in pairs(filter) do
				if not c:Equals(filterValue, binding[filterName]) then
					applies = false;
					break;
				end
			end
			if applies then
				result[index] = binding;
			end
		end
		return result;
	else
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS];
	end
end

function c:DeleteEventBinding(index)
	if index and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS][index]then
		local binding = c:Deepcopy(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS][index]);
		table.remove(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_EVENTS], index);

		-- notification
		local event = c:GetEvents()[binding[c.FIELD_TYPE]];
		if event and binding[c.FIELD_SUBTYPE] then
			event = event .. " (" .. c:GetZoneDisplayName(binding[c.FIELD_SUBTYPE]["mapID"]) .. ")";
		end
		local setName = binding[c.FIELD_NAME];
		if setName == c.KEYWORD_PREVIOUS then
			setName = c:GetText("[Previous set]");
		end
		c:NotifyInChat(string.format("%s %s -> %s", c:GetText("Removed event:"), event, setName));
	end
end

function c:BindingExists(binding)
	if binding[c.FIELD_TYPE] then
		local bindingClone = c:Deepcopy(binding);
		if bindingClone[c.FIELD_SUBSUBTYPE] then
			bindingClone[c.FIELD_SUBTYPE] = bindingClone[c.FIELD_SUBSUBTYPE];
			bindingClone[c.FIELD_SUBSUBTYPE] = nil;
		end
		bindingClone[c.FIELD_PVE] = nil;
		bindingClone[c.FIELD_PVP] = nil;

		for index, matchedBinding in pairs(c:LoadEventBindings(bindingClone)) do
			if (binding[c.FIELD_PVE] and matchedBinding[c.FIELD_PVE]) or (binding[c.FIELD_PVP] and matchedBinding[c.FIELD_PVP]) then
				return true;
			end
		end
	end
end

----- [ other ] -----

function c:SaveCloakAndHelmet(setName, notify)
	setName = setName or c:LoadCurrentSetName();
	if notify == nil then
		notify = true;
	end

	if setName then
		if ShowingHelm() ~= GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] = ShowingHelm();
			if notify then
				c:NotifyInChat(string.format("<%s> %s %s", setName, c:GetText("Saved helmet visibility:"), c:BoolToText(ShowingHelm(), c:GetText("visible"), c:GetText("not visible"))));
			end
		end

		if ShowingCloak() ~= GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] then
			GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] = ShowingCloak();
			if notify then
				c:NotifyInChat(string.format("<%s> %s %s", setName, c:GetText("Saved cloak visibility:"), c:BoolToText(ShowingCloak(), c:GetText("visible"), c:GetText("not visible"))));
			end
		end
	end
end

function c:LoadCloakAndHelmet(setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET] ~= nil then
			ShowHelm(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWHELMET]);
		end
		if GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK] ~= nil then
			ShowCloak(GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_SHOWCLOAK]);
		end
	end
end

function c:LoadPartialOption(setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_OPT_PARTIAL];
	end
end

function c:SavePartialOption(setName, value)
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_OPT_PARTIAL] = value;
	end
end

function c:LoadActionSlotsOption(setName)
	setName = setName or c:LoadCurrentSetName();
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_OPT_ACTIONSLOTS];
	end
end

function c:SaveActionSlotsOption(setName, value)
	if setName and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName] then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_SETS][setName][FIELD_OPT_ACTIONSLOTS] = value;
	end
end

function c:SaveLastBagLocation(itemString, bagId, slotId)
	if itemString then
		GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][itemString] = { [c.ITEM_BAG_ID] = bagId, [c.ITEM_SLOT_ID] = slotId };
	end
end

function c:LoadLastBagLocation(itemString)
	if itemString and GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][itemString] then
		return GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][itemString][c.ITEM_BAG_ID], GQ_DATA[c:GetRealmName()][c:GetCharName()][FIELD_INVENTORY][itemString][c.ITEM_SLOT_ID];
	end
end